from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import tensorflow as tf
tf.get_logger().setLevel('ERROR')
tf.compat.v1.disable_eager_execution()

from . import constraints
from . import functionals
from . import models
from . import utils

from .functionals.functional import Functional
from .functionals.variable import Variable
from .functionals.field import Field
from .functionals.parameter import Parameter
from .models.model import SciModel
from .constraints import Constraint, PDE, Data, Tie

# Also importable from root
from .utils.math import *
from .utils import math
from .utils.utilities import reset_session, clear_session
from .utils.utilities import set_random_seed
from .utils.utilities import set_floatx

# SciANN.
__author__ = "Ehsan Haghighat"
__email__ = "ehsanh@mit.edu"
__copyright__ = "Copyright 2019, Physics-Informed Deep Learning"
__credits__ = []
__url__ = "http://github.com/sciann/sciann]"
__license__ = "MIT"
__version__ = "0.5.3.4"
__cite__ = \
    '@misc{haghighat2019sciann, \n' +\
    '    title={SciANN: A Keras/Tensorflow wrapper for scientific computations and physics-informed deep learning using artificial neural networks}, \n' +\
    '    author={Ehsan Haghighat and Ruben Juanes}, \n' +\
    '    year={2020}, \n' +\
    '    eprint={2005.08803}, \n' +\
    '    archivePrefix={arXiv}, \n' +\
    '    primaryClass={cs.OH}, \n' +\
    '    url = {https://arxiv.org/abs/2005.08803}' +\
    '    howpublished={https://github.com/sciann/sciann.git}' +\
    '}'

# Import message.
_header = '---------------------- {} {} ----------------------'.format(str.upper(__name__), str(__version__))
_footer = len(_header)*'-'
__welcome__ = \
    '{} \n'.format(_header) +\
    'For details, check out our review paper and the documentation at: \n' +\
    ' +  "https://arxiv.org/abs/2005.08803", \n' +\
    ' +  "https://www.sciann.com". \n'
    # '{} \n'.format(__cite__) +\
    # _footer


import os
if 'SCIANN_WELCOME_MSG' in os.environ.keys() and \
        os.environ['SCIANN_WELCOME_MSG']=='-1':
    pass
else:
    print(__welcome__)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..utils import default_bias_initializer, default_kernel_initializer, default_constant_initializer
from tensorflow.python.keras.layers import Dense, SimpleRNN, LSTM, TimeDistributed
from tensorflow.python.keras.activations import linear

from ..utils import default_bias_initializer, default_kernel_initializer
from ..utils import default_regularizer
from ..utils import floatx, set_floatx


class RNNField(TimeDistributed):
    """ Configures the `LSTMField` class for the model outputs.

    # Arguments
        units: Positive integer.
            Dimension of the output of the network.
        name: String.
            Assigns a layer name for the output.
        activation: Callable.
            A callable object for the activation.
        kernel_initializer: Initializer for the kernel.
            Defaulted to a normal distribution.
        bias_initializer: Initializer for the bias.
            Defaulted to a normal distribution.
        kernel_regularizer: Regularizer for the kernel.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        bias_regularizer: Regularizer for the bias.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        trainable: Boolean to activate parameters of the network.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').

    # Raises

    """
    def __init__(self, units=1,
                 name=None,
                 rnn_type='SimpleRNN',
                 activation=linear,
                 kernel_initializer=default_kernel_initializer(),
                 recurrent_initializer=default_kernel_initializer(),
                 bias_initializer=default_bias_initializer(),
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 trainable=True,
                 dtype=None,):
        if not dtype:
            dtype = floatx()
        elif not dtype == floatx():
            set_floatx(dtype)

        assert isinstance(name, str), \
            "Please provide a string for field name. "
        assert callable(activation), \
            "Please provide a function handle for the activation. "

        # prepare initializers.
        if isinstance(kernel_initializer, (float, int)):
            kernel_initializer = default_constant_initializer(kernel_initializer)
        if isinstance(bias_initializer, (float, int)):
            bias_initializer = default_constant_initializer(bias_initializer)
        # prepare regularizers.
        kernel_regularizer = default_regularizer(kernel_regularizer)
        bias_regularizer = default_regularizer(bias_regularizer)

        if rnn_type == 'SimpleRNN':
            super(RNNField, self).__init__(
                SimpleRNN(
                    units=units,
                    activation=activation,
                    return_sequences=True,
                    kernel_initializer=kernel_initializer,
                    recurrent_initializer=recurrent_initializer,
                    bias_initializer=bias_initializer,
                    kernel_regularizer=kernel_regularizer,
                    recurrent_regularizer=recurrent_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    dtype=dtype,
                    unroll=True,
                    name=name
                )
            )
        elif rnn_type == 'LSTM':
            super(RNNField, self).__init__(
                LSTM(
                    units=units,
                    activation=activation,
                    return_sequences=True,
                    kernel_initializer=kernel_initializer,
                    recurrent_initializer=recurrent_initializer,
                    bias_initializer=bias_initializer,
                    kernel_regularizer=kernel_regularizer,
                    recurrent_regularizer=recurrent_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    dtype=dtype,
                    unroll=True,
                    name=name
                )
            )
        elif rnn_type == 'Dense':
            super(RNNField, self).__init__(
                Dense(
                    units=units,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    bias_initializer=bias_initializer,
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    dtype=dtype,
                    name=name
                )
            )
        else:
            raise NotImplementedError('Supported RNNType: (SimpleRNN, LSTM, Dense)')



""" PDE class to impose pde constraint.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .constraint import Constraint
from ..utils import is_functional
from ..utils import relu, sign, abs, tanh


class MinMax(Constraint):
    """MinMax weight constraint.

    Constrains the weights incident to each hidden unit
    to have values between a lower bound and an upper bound.

    # Arguments
        min_value: the minimum norm for the incoming weights.
        max_value: the maximum norm for the incoming weights.

    # Raises
        ValueError: 'cond' should be a functional object.
    """
    def __init__(self, cond, min_value=None, max_value=None, penalty=1.0, name="minmax"):
        if not is_functional(cond):
            raise ValueError(
                "Expected a Functional object, received a "
                "{} - {}".format(type(cond), cond)
            )
        if min_value is not None and max_value is not None and min_value > max_value:
            raise ValueError(
                "Check inputs: `min_value` should be smaller than `max_value`. "
            )
        try:
            delta = max_value - min_value
            const = 0.0
            if min_value is not None:
                const += (1.0 - sign(cond - min_value)) * abs(cond - min_value)
            if max_value is not None:
                const += (1.0 + sign(cond - max_value)) * abs(cond - max_value)
            const *= penalty
        except (ValueError, TypeError):
            assert False, 'Unexpected error - cannot evaluate the regularization. '

        super(MinMax, self).__init__(
            cond=const,
            name=name
        )



""" Utilities to process functionals.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import sciann


def is_functional(f):
    """ Checks whether `f` is a functional object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if functional.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, (sciann.Functional, sciann.functionals.RNNFunctional)):
        return True

    else:
        return False


def validate_functional(f):
    """ if `f` is not a functional object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if functional, False otherwise.

    # Raises
        ValueError: if the object is not a Functional object.

    """
    if isinstance(f, (sciann.Functional, sciann.functionals.rnn_functional.RNNFunctional)):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `functional` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors.'
        )


def is_constraint(f):
    """ Checks whether `f` is a `Constraint` object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Constraint.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, sciann.Constraint):
        return True

    else:
        return False


def validate_constraint(f):
    """ if `f` is not a Constraint object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Constraint, False otherwise.

    # Raises
        ValueError: if the object is not a Constraint object.

    """
    if isinstance(f, sciann.Constraint):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `Constraint` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors '
            'or layers. '
        )


def is_parameter(f):
    """ Checks whether `f` is a parameter object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if a parameter.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, sciann.Parameter):
        return True

    else:
        return False


def validate_parameter(f):
    """ if `f` is not a parameter object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if parameter, False otherwise.

    # Raises
        ValueError: if the object is not a Parameter object.

    """
    if isinstance(f, sciann.Parameter):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `parameter` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors.'
        )


def is_field(f):
    """ Checks whether `f` is a `Field` object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Field.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, (sciann.Field, sciann.functionals.RNNField)):
        return True

    else:
        return False


def validate_field(f):
    """ if `f` is not a Field object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Field, False otherwise.

    # Raises
        ValueError: if the object is not a Field object.

    """
    if isinstance(f, (sciann.Field, sciann.functionals.RNNField)):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `Field` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors '
            'or layers. '
        )


def is_variable(f):
    """ Checks whether `f` is a `Variable` object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Variable.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, (sciann.Variable, sciann.functionals.RadialBasis, sciann.functionals.RNNVariable)):
        return True

    else:
        return False


def validate_variable(f):
    """ if `f` is not a Variable object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if Variable, False otherwise.

    # Raises
        ValueError: if the object is not a Variable object.

    """
    if isinstance(f, (sciann.Variable, sciann.functionals.RadialBasis, sciann.functionals.RNNVariable)):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `Variable` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors '
            'or layers. '
        )


def is_scimodel(f):
    """ Checks whether `f` is a `SciModel` object.

    # Arguments
        f: an object to be tested.

    # Returns
        True if SciModel.

    # Raises
        ValueError: if the object cannot be tested with `isinstance`.

    """
    if isinstance(f, sciann.SciModel):
        return True

    else:
        return False


def validate_scimodel(f):
    """ if `f` is not a SciModel object, raises value error.

    # Arguments
        f: an object to be tested.

    # Returns
        True if SciModel, False otherwise.

    # Raises
        ValueError: if the object is not a SciModel object.

    """
    if isinstance(f, sciann.SciModel):
        return True

    else:
        raise ValueError(
            'These operations can only be applied to the `SciModel` object. '
            'Use `Keras` or `TensorFlow` functions when applying to tensors '
            'or layers. '
        )



'''
# Curve fitting in 1D

Here, a 1D curve fitting example is explored. Imagine, a synthetic data
generated from \\\( \sin(x) \\\) over the range of \\\( [0, 2\pi] \\\).

To train a neural network model on this curve, you should first define a `Variable`.

A neural network with three layers, each containing 10 neurons, and with `tanh` activation function is then generated
using the `Functional` class.

The target is imposed on the output using the `Data` class from `Constraint`, and passed to the `SciModel` to form a
SciANN model.
'''

import numpy as np
from sciann import Variable, Functional, SciModel, Parameter
from sciann.constraints import Data, MinMax
from sciann.utils.math import diff
import sciann as sn

sn.set_random_seed(1234)
# Synthetic data generated from sin function over [0, 2pi]
x_true = np.linspace(0, np.pi*2, 10000)
y_true = np.sin(x_true)

# The network inputs should be defined with Variable.
x = Variable('x', dtype='float64')

# Each network is defined by Functional.
y = Functional('y', x, [10, 10, 10], activation='tanh')

d = Parameter(10.0, inputs=x)

# Define the target (output) of your model.
c1 = Data(y)

L = d*diff(y, x, order=2) + y

# The model is formed with input `x` and condition `c1`.
model = SciModel(x, [c1, sn.PDE(L)])

# Training: .train runs the optimization and finds the parameters.
history = model.train(
    x_true,
    [y_true, 'zeros'],
    batch_size=32,
    epochs=100,
    adaptive_weights=True
)

# used to evaluate the model after the training.
y_pred = y.eval(model, x_true)
d_pred = d.eval(model, x_true)




# -*- coding: utf-8 -*-
'''
General documentation architecture:

Home
Index

- Getting started
    Guide to SciANN model
    Guide to Functional
    FAQ

- SciModels
    About SciANN models
        explain when one should use Sequential or functional API
        explain compilation step
        explain weight saving, weight loading
        explain serialization, deserialization
    Sequential
    Model (functional API)

- Functionals
    About SciANN functionals
        explain common functional operations: get_weights, set_weights, get_config
        explain input_shape
        explain usage on non-Keras tensors
    Variables
    Fields
    Functionals 

Operations
Constraints 

'''
from sciann import utils
from sciann import Functional
from sciann import Variable
from sciann import Field
from sciann import SciModel
from sciann import constraints
from sciann import utils
from sciann import Parameter


EXCLUDE = {
    'Constraint'
}

PAGES = [
    {
        'page': 'scimodels.md',
        'classes': [
            SciModel
        ],
        'methods': [
            SciModel.train,
            SciModel.predict,
            SciModel.loss_functions,
        ]
    },
    {
        'page': 'functionals.md',
        'classes': [
            Functional,
            Variable,
            Field,
            Parameter,
        ],
        'methods': [
            Functional.eval,
            Functional.get_weights,
            Functional.set_weights,
            Functional.reinitialize_weights,
            Functional.count_params,
            Functional.set_trainable,
            Functional.split,
        ]
    },
    {
        'page': 'variables.md',
        'classes': [
            Variable
        ]
    },
    {
        'page': 'fields.md',
        'classes': [
            Field
        ]
    },
    {
        'page': 'constraints.md',
        'classes': [
            constraints.Data,
            constraints.PDE,
            constraints.Tie
        ]
    },
    {
        'page': 'utils.md',
        'methods': [
            utils.math.grad,
            utils.math.diag_grad,
            utils.math.div,
            utils.math.radial_basis,
            utils.math.sin,
            utils.math.asin,
            utils.math.cos,
            utils.math.acos,
            utils.math.tan,
            utils.math.atan,
            utils.math.tanh,
            utils.math.exp,
            utils.math.pow,
            utils.math.add,
            utils.math.sub,
            utils.math.mul,
            utils.math.div,
        ],
    },
]

ROOT = 'https://sciann.com/'

template_np_implementation = """# Numpy implementation

    ```python
{{code}}
    ```
"""

template_hidden_np_implementation = """# Numpy implementation

    <details>
    <summary>Show the Numpy implementation</summary>

    ```python
{{code}}
    ```

    </details>
"""



""" Functional class for SciANN.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.python.keras.backend as K
graph_unique_name = K.get_graph().unique_name

from tensorflow.python.keras.layers import Dense, LSTM, SimpleRNN
from tensorflow.python.keras.layers import Activation
from tensorflow.python.keras.layers import Concatenate

from ..utils import to_list, unpack_singleton, is_same_tensor, unique_tensors
from ..utils import default_weight_initializer
from ..utils import default_regularizer
from ..utils import to_list, unpack_singleton
from ..utils import validations, getitem
from ..utils import math
from ..utils.activations import SciActivation, get_activation
from ..utils import prepare_default_activations_and_initializers

from .rnn_field import RNNField


class RNNFunctional(object):
    """ Configures the LSTMFunctional object (Recurrent Neural Network).

    # Arguments
        fields: String or Field.
            [Sub-]Network outputs.
            It can be of type `String` - Associated fields will be created internally.
            It can be of type `Field` or `Functional`
        variables: Variable.
            [Sub-]Network inputs.
            It can be of type `Variable` or other Functional objects.
        hidden_layers: A list indicating neurons in the hidden layers.
            e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
        activation: Activation function for the hidden layers.
            Last layer will have a linear output.
        output_activation: defaulted to "linear".
            Activation function to be applied to the network output.
        rnn_type: currently, `SimpleRNN` and `LSTM` are accepted.
            Defaulted to `SimpleRNN`.
            Check `Keras` documentation for additional information.
        kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
        recurrent_initializer: Initializer of the `Recurrent Kernel`, from `k.initializers`.
        bias_initializer: Initializer of the `Bias`, from `k.initializers`.
        kernel_regularizer: Regularizer for the kernel.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        bias_regularizer: Regularizer for the bias.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').
            Note: Only network inputs should be set.
        trainable: Boolean.
            False if network is not trainable, True otherwise.
            Default value is True.

    # Raises
        ValueError:
        TypeError:
    """
    def __init__(self,
                 fields=None,
                 variables=None,
                 hidden_layers=None,
                 activation="tanh",
                 output_activation="linear",
                 rnn_type="SimpleRNN",
                 recurrent_activation="tanh",
                 kernel_initializer=None,
                 recurrent_initializer=None,
                 bias_initializer=None,
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 dtype=None,
                 trainable=True,
                 **kwargs):
        # check data-type.
        if dtype is None:
            dtype = K.floatx()
        elif not K.floatx() == dtype:
            K.set_floatx(dtype)
        # prepare hidden layers.
        if hidden_layers is None:
            hidden_layers = []
        else:
            hidden_layers = to_list(hidden_layers)
        # check for copy constructor.
        if all([x in kwargs for x in ('inputs', 'outputs', 'layers')]):
            self._inputs = kwargs['inputs'].copy()
            self._outputs = kwargs['outputs'].copy()
            self._layers = kwargs['layers'].copy()
            self._set_model()
            return
        # prepare kernel initializers.
        activations, def_biasinit, def_kerinit = \
            prepare_default_activations_and_initializers(
            len(hidden_layers) * [activation] + [output_activation]
        )
        # prepare kernel initializer.
        if kernel_initializer is None:
            kernel_initializer = def_kerinit
        elif isinstance(kernel_initializer, (float, int)):
            kernel_initializer = default_weight_initializer(
                len(hidden_layers) * [activation] + [output_activation],
                'constant',
                scale=kernel_initializer
            )
        else:
            kernel_initializer = [kernel_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
        # prepare recurrent initializer.
        if recurrent_initializer is None:
            recurrent_initializer = def_kerinit
        elif isinstance(recurrent_initializer, (float, int)):
            recurrent_initializer = default_weight_initializer(
                len(hidden_layers) * [activation] + [output_activation],
                'constant',
                scale=recurrent_initializer
            )
        else:
            recurrent_initializer = [recurrent_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
        # prepare bias initializers.
        if bias_initializer is None:
            bias_initializer = def_biasinit
        elif isinstance(bias_initializer, (float, int)):
            bias_initializer = default_weight_initializer(
                len(hidden_layers) * [activation] + [output_activation],
                'constant',
                scale=bias_initializer
            )
        else:
            bias_initializer = [bias_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
        # prepare regularizers.
        kernel_regularizer = default_regularizer(kernel_regularizer)
        recurrent_regularizer = default_regularizer(recurrent_regularizer)
        bias_regularizer = default_regularizer(bias_regularizer)
        # prepares fields.
        fields = to_list(fields)
        if all([isinstance(fld, str) for fld in fields]):
            output_fields = [
                RNNField(
                    name=fld,
                    rnn_type='Dense',
                    dtype=dtype,
                    kernel_initializer=kernel_initializer[-1],
                    recurrent_initializer=recurrent_initializer[-1],
                    bias_initializer=bias_initializer[-1],
                    kernel_regularizer=kernel_regularizer,
                    recurrent_regularizer=recurrent_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                )
                for fld in fields
            ]
        elif all([validations.is_field(fld) for fld in fields]):
            output_fields = fields
        else:
            raise TypeError(
                'Please provide a "list" of field names of'
                + ' type "String" or "Field" objects.'
            )
        # prepare inputs/outputs/layers.
        inputs = []
        layers = []
        variables = to_list(variables)
        if all([isinstance(var, RNNFunctional) for var in variables]):
            for var in variables:
                inputs += var.outputs
            # for var in variables:
            #     for lay in var.layers:
            #         layers.append(lay)
        else:
            raise TypeError(
                "Input error: Please provide a `list` of `Functional`s. \n"
                "Provided - {}".format(variables)
            )
        # Check and convert activation functions to proper format.
        assert not isinstance(activation, list), \
            'Expected an activation function name not a "list". '
        afunc = get_activation(activation)

        # Input layers.
        if len(inputs) == 1:
            net_input = inputs[0]
        else:
            layer = Concatenate(name=graph_unique_name('conct'))
            net_input = layer(inputs)

        # Define the networks.
        net = [net_input]

        # Adding hidden layers
        for nLay, nNeuron in enumerate(hidden_layers):
            if nLay < 1000:
                # First layer starts with RNN.
                if rnn_type=='LSTM':
                    layer = LSTM(
                        nNeuron,
                        return_sequences=True,
                        activation=None,
                        recurrent_activation=recurrent_activation,
                        kernel_initializer=kernel_initializer[nLay],
                        recurrent_initializer=recurrent_initializer[nLay],
                        bias_initializer=bias_initializer[nLay],
                        kernel_regularizer=kernel_regularizer,
                        recurrent_regularizer=recurrent_regularizer,
                        bias_regularizer=bias_regularizer,
                        trainable=trainable,
                        dtype=dtype,
                        unroll=True,
                        name=graph_unique_name("LSTM{:d}b_".format(nNeuron))
                    )
                elif rnn_type=='SimpleRNN':
                    layer = SimpleRNN(
                        nNeuron,
                        activation=None,
                        return_sequences=True,
                        kernel_initializer=kernel_initializer[nLay],
                        recurrent_initializer=recurrent_initializer[nLay],
                        bias_initializer=bias_initializer[nLay],
                        kernel_regularizer=kernel_regularizer,
                        recurrent_regularizer=recurrent_regularizer,
                        bias_regularizer=bias_regularizer,
                        trainable=trainable,
                        dtype=dtype,
                        unroll=True,
                        name=graph_unique_name("SRNN{:d}b_".format(nNeuron))
                    )
                else:
                    raise ValueError(
                        'Invalid entry for `rnn_type` -- '
                        'accepts from (`SimpleRNN`, `LSTM`).'
                    )
            else:
                # Add the dense layer.
                layer = Dense(
                    nNeuron,
                    kernel_initializer=kernel_initializer[nLay],
                    bias_initializer=bias_initializer[nLay],
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    dtype=dtype,
                    name=graph_unique_name("D{:d}b".format(nNeuron))
                )
            layers.append(layer)
            net[-1] = layer(net[-1])
            # Apply the activation.
            if afunc.__name__ != 'linear':
                layer = activations[nLay]
                layers.append(layer)
                net[-1] = layer(net[-1])

        # store output layers.
        for out in output_fields:
            layers.append(out)

        # Assign to the output variable
        if len(net) == 1:
            net_output = net[0]
        else:
            raise ValueError("Legacy for Enrichment: Must be updated. ")
            layer = Concatenate(name=graph_unique_name("{}_".format("conct")))
            net_output = layer(net)

        # check output activation functions.
        output_func = get_activation(output_activation)
        # Define the final outputs of each network
        outputs = []
        for out in output_fields:
            # add the activation on the output.
            if output_func.__name__ != 'linear':
                layer = activations[-1]
                layers.append(layer)
                outputs.append(layer(out(net_output)))
            else:
                outputs.append(out(net_output))

        self._inputs = inputs
        self._outputs = outputs
        self._layers = layers
        self._set_model()

    def eval(self, *kwargs):
        """ Evaluates the functional object for a given input.

        # Arguments
            (SciModel, Xs):
                Evalutes the functional object from the beginning
                    of the graph defined with SciModel.
                    The Xs should match those of SciModel.

            (Xs):
                Evaluates the functional object from inputs of the functional.
                    Xs should match those of inputs to the functional.

        # Returns
            Numpy array of dimensions of network outputs.

        # Raises
            ValueError:
            TypeError:
        """
        if len(kwargs) == 1:
            model = self.model
            # read data.
            mesh = kwargs[0]
        elif len(kwargs) == 2:
            if validations.is_scimodel(kwargs[0]):
                model = K.function(kwargs[0].model.inputs, self.outputs)
            else:
                raise ValueError(
                    'Expected a SciModel object for the first arg. '
                )
            mesh = kwargs[1]
        else:
            raise NotImplemented()
        x_pred = to_list(mesh.copy())
        # To have unified output for postprocessing - limitted support.
        shape_default = x_pred[0].shape if all([x.shape == x_pred[0].shape for x in x_pred]) else None
        # prepare X,Y data.
        for i, (x, xt) in enumerate(zip(x_pred, model.inputs)):
            x_shape = tuple(xt.get_shape().as_list())
            if x.shape != x_shape:
                try:
                    x_pred[i] = x.reshape((-1,) + x_shape[1:])
                except:
                    print(
                        'Could not automatically convert the inputs to be '
                        'of the same size as the expected input tensors. '
                        'Please provide inputs of the same dimension as the `Variables`. '
                    )
                    assert False

        y_pred = to_list(model(x_pred))

        if shape_default is not None:
            try:
                y_pred = [y.reshape(shape_default) for y in y_pred]
            except:
                print("Input and output dimensions need re-adjustment for post-processing.")

        return unpack_singleton(y_pred)

    @property
    def layers(self):
        return self._layers

    @layers.setter
    def layers(self, value):
        self._layers = value

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, value):
        self._inputs = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def model(self):
        self._set_model()
        return self._model

    def _set_model(self):
        if hasattr(self, '_model'):
            if is_same_tensor(self._inputs, self._model.inputs) and \
               is_same_tensor(self._outputs, self._model.outputs):
               return
        self._model = K.function(
            unique_tensors(self._inputs),
            self._outputs
        )

    def get_weights(self, at_layer=None):
        return [l.get_weights() for l in self.layers]

    def set_weights(self, weights):
        try:
            for l, w in zip(self.layers, weights):
                l.set_weights(w)
        except:
            raise ValueError(
                'Provide data exactly the same as .get_weights() outputs. '
            )

    def count_params(self):
        return sum([l.count_params() for l in self.layers])

    def copy(self):
        return RNNFunctional(
            inputs=self.inputs,
            outputs=self.outputs,
            layers=self.layers
        )

    def append_to_layers(self, layers):
        if self.layers is not layers:
            cl = [x.name for x in self.layers]
            for x in layers:
                if not x.name in cl:
                    self.layers += [x]

    def append_to_inputs(self, inputs):
        if self.inputs is not inputs:
            cl = [x.name for x in self.inputs]
            for x in inputs:
                if not x.name in cl:
                    self.inputs.append(x)

    def append_to_outputs(self, outputs):
        self._outputs += to_list(outputs)

    def set_trainable(self, val):
        if isinstance(val, bool):
            for l in self._layers:
                l.trainable = val
        else:
            raise ValueError('Expected a boolean value: True or False')
        return self

    def reinitialize_weights(self):
        for lay in self.layers:
            if hasattr(lay, 'kernel_initializer') and lay.kernel is not None:
                K.set_value(lay.kernel, lay.kernel_initializer(lay.kernel.shape))
            if hasattr(lay, 'bias_initializer') and lay.bias is not None:
                K.set_value(lay.bias, lay.bias_initializer(lay.bias.shape))
        return self

    def split(self):
        """ In the case of `Functional` with multiple outputs,
            you can split the outputs and get an associated functional.

        # Returns
            (f1, f2, ...): Tuple of splitted `Functional` objects
                associated to cheach outputs.
        """
        if len(self._outputs)==1:
            return self
        fs = ()
        # total number of outputs to get splitted.
        nr = len(self._outputs)
        # associated to each output, there is a layer to be splitted.
        lays = self.layers[:-nr]
        for out, lay in zip(self._outputs, self._layers[-nr:]):
            # copy constructor for functional.
            f = RNNFunctional(
                inputs = to_list(self.inputs),
                outputs = to_list(out),
                layers = lays + to_list(lay)
            )
            fs += (f,)
        return fs

    def __call__(self):
        return self.outputs

    def __pos__(self):
        return self

    def __neg__(self):
        return self*-1.0

    def __add__(self, other):
        return math.add(self, other)

    def __radd__(self, other):
        return math.radd(self, other)

    def __sub__(self, other):
        return math.sub(self, other)

    def __rsub__(self, other):
        return math.rsub(self, other)

    def __mul__(self, other):
        return math.mul(self, other)

    def __rmul__(self, other):
        return math.rmul(self, other)

    def __truediv__(self, other):
        return math.div(self, other)

    def __rtruediv__(self, other):
        return math.rdiv(self, other)

    def __pow__(self, power):
        return math.pow(self, power)

    def __getitem__(self, item):
        return getitem(self, item)

    def diff(self, *args, **kwargs):
        return math.diff(self, *args, **kwargs)

    @classmethod
    def get_class(cls):
        return RNNFunctional


# Scipy wrapper for keras
# originally developed by:
#   https://github.com/rohitgr7/keras-opt
#   addopted by Ehsan Haghighat

import numpy as np
from scipy.optimize import minimize
from tensorflow import keras
from tensorflow.keras import backend as K  # pylint: disable=import-error
from tensorflow.python.keras.callbacks import BaseLogger, CallbackList, History  # pylint: disable=no-name-in-module
from tensorflow.python.keras.optimizer_v2 import optimizer_v2   # pylint: disable=no-name-in-module
from tensorflow.python.ops import variables as tf_variable

from tensorflow.python.keras.utils.generic_utils import Progbar

# from tqdm import trange, tqdm_notebook


class GradientObserver(optimizer_v2.OptimizerV2):
    """
    Implements the Keras Optimizer interface in order to accumulate gradients for
    each mini batch. Gradients are then read at the end of the epoch by the ScipyOptimizer. 
    """

    def __init__(self, learning_rate=0.001, method='L-BFGS-B', **kwargs):
        super(GradientObserver, self).__init__('GradientObserver')
        self._learning_rate = tf_variable.Variable(kwargs.get('lr', learning_rate), 'float32')
        self._method = method.lower().split("scipy-")[-1]
        self._vars = []
        self._grads = {}

    @property
    def learning_rate(self):
        return self._learning_rate

    @property
    def lr(self):
        return self._learning_rate

    @property
    def method(self):
        return self._method

    def get_updates(self, loss, params):
        """
        Build the graph nodes that accumulate gradients.
        """
        self.updates = []
        grads = self.get_gradients(loss, params)
        for param, grad in zip(params, grads):
            shape = K.int_shape(param)
            var = K.zeros(shape)
            self._vars.append(var)
            self.updates.append(K.update_add(var, grad))
        return self.updates

    def get_gradient_values(self):
        """
        Read gradient values (at epoch end).
        """
        values = []
        for g in self._grads.values():
            values.append(K.eval(g))
        for v in self._vars:
            values.append(K.eval(v))
        return values

    def clear(self):
        """
        Clear gradient values (used at epoch start)
        """
        self._grads = {}
        for var in self._vars:
            K.set_value(var, np.zeros(var.shape))

    def _create_slots(self, var_list):
        pass

    def _resource_apply_dense(self, grad, var, **apply_kwargs):
        self._grads[var.name] = grad

    def _resource_apply_sparse(self, grad, handle, indices, apply_state):
        if handle.name in self._grads:
            dense_grad = self._grads[handle.name]
        else:
            dense_grad = np.zeros(handle.shape.as_list())
            self._grads[handle.name] = dense_grad

        for i, idx in enumerate(indices.numpy()):
            dense_grad[idx] += grad[i]

    def get_config(self):
        config = super(GradientObserver, self).get_config()
        return config


class GeneratorWrapper(keras.utils.Sequence):
    """
    Converts fit() into fit_generator() interface.
    """

    def __init__(self, inputs, outputs, sample_weights, batch_size, shuffle):
        self._inputs = inputs
        self._outputs = outputs
        self._sample_weights = sample_weights
        self._size = inputs[0].shape[0]
        self._batch_size = batch_size
        self._num_batches = int((self._size-1)/batch_size) + 1
        self._shuffle = shuffle
        self._ids = np.arange(0, self._size)
        print("\nTotal samples: {} ".format(self._size))
        print("Batch size: {} ".format(min(self._batch_size, self._size)))
        print("Total batches: {} \n".format(self._num_batches))

    def __len__(self):
        return self._num_batches

    def __getitem__(self, index):
        if self._num_batches > 1 and index == 0 and self._shuffle:
            self._ids = np.random.choice(self._size, self._size, replace=False)
        start = index * self._batch_size
        end = min(start + self._batch_size, self._size)
        ids = self._ids[start: end]
        inputs = [v[ids, :] for v in self._inputs]
        outputs = [v[ids, :] for v in self._outputs]
        sample_weights = [v[ids] for v in self._sample_weights]
        return inputs, outputs, sample_weights


class ScipyOptimizer(object):
    """
    Invokes the underlying model in order to obtain the cost and gradients for the function
    being optimized.
    """

    def __init__(self, model):
        self._model = model
        self._layers = [layer for layer in model._layers if layer.weights]
        self._weights_size = 0
        for layer in self._layers:
            for w in layer.weights:
                if not w.trainable:
                    continue
                self._weights_size += np.prod(w.shape)

    def _update_weights(self, x):
        x_offset = 0
        for layer in self._layers:
            w_list = []
            w_trainable = [w.trainable for w in layer.weights]
            batch_update = False not in w_trainable
            for w in layer.weights:
                if not w.trainable:
                    continue
                shape = w.get_shape()
                w_size = np.prod(shape)
                value = np.array(x[x_offset:x_offset+w_size]).reshape(shape)
                if batch_update:
                    w_list.append(value)
                else:
                    K.set_value(w, value)
                x_offset += w_size
            if batch_update:
                layer.set_weights(w_list)
        assert x_offset == self._weights_size

    def _collect_weights(self):
        x_values = np.empty(self._weights_size)
        x_offset = 0
        for layer in self._layers:
            w_trainable = [w.trainable for w in layer.weights]
            for var, trainable in zip(layer.get_weights(), w_trainable):
                if not trainable:
                    continue
                w_size = var.size
                x_values[x_offset:x_offset+w_size] = var.reshape(-1)
                x_offset += w_size
        assert x_offset == self._weights_size
        return x_values

    def _fun_generator(self, x, generator, state):
        self._model.optimizer.clear()
        self._update_weights(x)
        callbacks = state['callbacks']

        if not state['in_epoch']:
            callbacks.on_epoch_begin(state['epoch'])
            state['in_epoch'] = True

        cost_sum = 0
        iterator = range(len(generator))
        if state['verbose'] == 1 and len(generator) > 1:
            # iterator = trange(len(generator))
            progress_bar = Progbar(len(generator))
        else:
            # iterator = range(len(generator))
            progress_bar = False

        state['epoch_logs'] = {}
        epoch_logs = state['epoch_logs']

        for batch_index in iterator:
            inputs, outputs, sample_weights = generator[batch_index]
            if isinstance(inputs, list):
                isize = inputs[0].shape[0]
            else:
                isize = inputs.shape[0]
            batch_logs = {'batch': batch_index, 'size': isize}
            callbacks.on_batch_begin(batch_index, batch_logs)
            outs = self._model.train_on_batch(inputs, outputs, sample_weights)
            if not isinstance(outs, list):
                outs = [outs]
            for lbl, v in zip(self._model.metrics_names, outs):
                batch_logs[lbl] = v
                epoch_logs[lbl] = epoch_logs.get(lbl, 0.0) + v
            callbacks.on_batch_end(batch_index, batch_logs)
            batch_cost = batch_logs['loss']
            if progress_bar:
                progress_bar.update(batch_index + 1)
                # iterator.set_postfix(cost=batch_cost)

            cost_sum += batch_cost

        generator.on_epoch_end()

        epoch_logs = state['epoch_logs']
        if state['verbose'] > 0:
            print('itr:', state['epoch'],
                  ', '.join(['{0}: {1:.4e}'.format(k, v) for k, v in epoch_logs.items()]))
        elif state['verbose'] < 0:
            print('itr:', state['epoch'], ', loss: {0:.6e}'.format(epoch_logs['loss']))

        # average the metrics
        for lbl in self._model.metrics_names:
            epoch_logs[lbl] = epoch_logs.get(lbl) / len(iterator)

        cost = cost_sum / len(iterator)

        gradients = self._model.optimizer.get_gradient_values()
        x_grad = np.empty(x.shape)
        x_offset = 0
        for grad in gradients:
            w_size = grad.size
            x_grad[x_offset:x_offset + w_size] = grad.reshape(-1)
            x_offset += w_size
        assert x_offset == self._weights_size
        self._cost = cost
        self._gradients = x_grad
        return cost, x_grad

    def _validate(self, x, val_generator, state):
        # TODO: weight update should be optimized in the most common case
        self._update_weights(x)
        # test callback are in the newer version of the CallbackList API.
        # callbacks = state['callbacks']
        epoch_logs = state['epoch_logs']

        # callbacks.on_test_begin()

        val_outs = [0] * len(self._model.metrics_names)
        n_steps = len(val_generator)
        for batch_index in range(n_steps):
            inputs, outputs = val_generator[batch_index]
            batch_logs = {'batch': batch_index, 'size': inputs.shape[0]}

            # callbacks.on_test_batch_begin(batch_index, batch_logs)
            batch_outs = self._model.test_on_batch(inputs, outputs)
            if not isinstance(batch_outs, list):
                batch_outs = [batch_outs]
            for l, o in zip(self._model.metrics_names, batch_outs):
                batch_logs[l] = o
            # callbacks.on_test_batch_end(batch_index, batch_logs)
            for i, batch_out in enumerate(batch_outs):
                val_outs[i] += batch_out

        for l, o in zip(self._model.metrics_names, val_outs):
            o /= n_steps
            epoch_logs['val_' + l] = o

        # callbacks.on_test_end()

    def fit(self, inputs, outputs, sample_weight, batch_size=32, shuffle=True, **kwargs):
        return self.fit_generator(GeneratorWrapper(inputs, outputs, sample_weight, batch_size, shuffle), **kwargs)

    def fit_generator(self, generator, epochs=1,
                      validation_data=None,
                      callbacks=None,
                      verbose=True):
        method = self._model.optimizer.method
        x0 = self._collect_weights()
        history = History()
        _callbacks = [BaseLogger(stateful_metrics=self._model.metrics_names)]
        _callbacks += (callbacks or []) + [history]
        callback_list = CallbackList(_callbacks)
        callback_list.set_model(self._model)
        callback_list.set_params({
            'epochs': epochs,
            'verbose': False,
            'metrics': list(self._model.metrics_names),
        })
        state = {
            'epoch': 0,
            'verbose': verbose,
            'callbacks': callback_list,
            'in_epoch': False,
            'epoch_logs': {},
        }
        min_options = {
            'maxiter': epochs,
        }

        val_generator = None
        if validation_data is not None:
            if isinstance(validation_data, keras.utils.Sequence):
                val_generator = validation_data
            elif isinstance(validation_data, tuple) and len(validation_data) == 2:
                val_generator = GeneratorWrapper(*validation_data)

        def on_iteration_end(xk):
            cb = state['callbacks']
            if val_generator is not None:
                self._validate(xk, val_generator, state)
            cb.on_epoch_end(state['epoch'], state['epoch_logs'])
            # if state['verbose']:
            #     epoch_logs = state['epoch_logs']
            #     print('epoch: ', state['epoch'],
            #           ', '.join([' {0}: {1:.3e}'.format(k, v) for k, v in epoch_logs.items()]))
            state['epoch'] += 1
            state['in_epoch'] = False
            state['epoch_logs'] = {}

        callback_list.on_train_begin()
        result = minimize(
            self._fun_generator, x0, method=method, jac=True, options=min_options,
            callback=on_iteration_end, args=(generator, state))
        self._update_weights(result['x'])
        callback_list.on_train_end()
        return history



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.keras.layers import InputLayer
from ..utils import floatx, set_floatx, to_list
from .functional import Functional


class Variable(Functional):
    """ Configures the `Variable` object for the network's input.

    # Arguments
        name: String.
            Required as derivatives work only with layer names.
        units: Int.
            Number of feature of input var.
        tensor: Tensorflow `Tensor`.
            Can be pass as the input path.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').

    # Raises

    """
    def __init__(self,
                 name=None,
                 units=1,
                 tensor=None,
                 dtype=None):

        if not dtype:
            dtype = floatx()
        elif not dtype == floatx():
            set_floatx(dtype)

        layer = InputLayer(
            # batch_input_shape=(None, 1) if units == 1 else (None, units, 1),
            batch_input_shape=(None, units),
            input_tensor=tensor,
            name=name,
            dtype=dtype
        )

        super(Variable, self).__init__(
            layers=to_list(layer),
            inputs=to_list(layer.input),
            outputs=to_list(layer.output)
        )

    @classmethod
    def get_class(cls):
        return Functional




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from . import constraint
from . import data
from . import pde
from . import tie

from .constraint import Constraint
from .data import Data
from .tie import Tie
from .pde import PDE
from .minmax import MinMax



""" Built-in utilities to process inputs.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from numpy import pi

import tensorflow as tf
import tensorflow.python.keras as k
import tensorflow.python.keras.backend as K

# interface for some keras features to be acessible across sciann.
from tensorflow.python.keras.backend import is_keras_tensor as is_tensor
from tensorflow.python.keras.backend import floatx
from tensorflow.python.keras.backend import set_floatx
from tensorflow.python.keras.utils.vis_utils import model_to_dot
from tensorflow.python.keras.utils.vis_utils import plot_model
from tensorflow.python.keras.initializers import RandomUniform as default_bias_initializer
from tensorflow.python.keras.initializers import GlorotNormal as default_kernel_initializer
from tensorflow.python.keras.initializers import Constant as default_constant_initializer
from tensorflow.python.keras.initializers import VarianceScaling
from tensorflow.python.keras.regularizers import l1_l2

from .initializers import SciKernelInitializer as KInitializer
from .initializers import SciBiasInitializer as BInitializer
from .activations import get_activation, SciActivation, SciActivationLayer


def _is_tf_1():
    return tf.__version__.startswith('1.')


def set_random_seed(val=1234):
    """ Set random seed for reproducibility.

    # Arguments
        val: A seed value..

    """
    np.random.seed(val)
    if _is_tf_1():
        tf.set_random_seed(val)
    else:
        tf.random.set_seed(val)


def reset_session():
    """ Clear keras and tensorflow sessions.
    """
    if _is_tf_1():
        K.clear_session()
    else:
        tf.keras.backend.clear_session()


clear_session = reset_session


def is_same_tensor(x, y):
    if len(to_list(x)) != len(to_list(y)):
        return False
    else:
        res = []
        for xi, yi in zip(to_list(x), to_list(y)):
            res.append(xi.name == yi.name)
        return all(res)


def unique_tensors(Xs):
    if len(Xs) > 1:
        ux, uids = np.unique([x.name for x in Xs], return_index=True)
        uids = sorted(uids)
        return [Xs[i] for i in uids]
    else:
        return Xs


def default_regularizer(*args, **kwargs):
    l1, l2 = 0.0, 0.0
    if (len(args) == 0 and len(kwargs) == 0) or args[0] is None:
        return None
    elif len(args) == 1:
        if isinstance(args[0], (float, int)):
            l1 = 0.0
            l2 = args[0]
        elif isinstance(args[0], list):
            l1 = args[0][0]
            l2 = args[0][1]
        elif isinstance(args[0], dict):
            l1 = 0.0 if 'l1' not in args[0] else args[0]['l1']
            l2 = 0.0 if 'l2' not in args[0] else args[0]['l2']
    elif len(args) == 2:
        l1 = args[0]
        l2 = args[1]
    elif len(kwargs) > 0:
        l1 = 0.0 if 'l1' not in kwargs else kwargs['l1']
        l2 = 0.0 if 'l2' not in kwargs else kwargs['l2']
    else:
        raise ValueError('Unrecognized entry - input regularization values for l1 and l2.')
    # print("regularization is used with l1={} and l2={}".format(l1, l2))
    return l1_l2(l1=l1, l2=l2)


def default_weight_initializer(actf='linear', distribution='uniform', mode='fan_in', scale=None):
    inz = []
    for i, af in enumerate(to_list(actf)):
        if distribution in ('uniform', 'normal'):
            tp = VarianceScaling(
                scale=eval_default_scale_factor(af, i) if scale is None else scale,
                mode=mode, distribution=distribution
            )
        elif distribution in ('constant',):
            tp = default_constant_initializer(0.0 if scale is None else scale)
        else:
            raise ValueError('Undefined distribution: pick from ("uniform", "normal", "constant").')
        inz.append(tp)
    return inz


def eval_default_scale_factor(actf, lay):
    if actf in ('linear', 'relu'):
        return 2.0
    elif actf in ('tanh', 'sigmoid'):
        return 1.0 if lay > 0 else 1.0
    elif actf in ('sin', 'cos'):
        return 2.0 if lay > 0 else 2.0 #*30.0
    else:
        return 1.0


def prepare_default_activations_and_initializers(actfs, seed=None):
    activations = []
    bias_initializer = []
    kernel_initializer = []
    for lay, actf in enumerate(to_list(actfs)):
        if isinstance(actf, str):
            lay_actf = actf.lower().split('l-')
        else:
            raise ValueError('expected a string for actf.')
        bias_initializer.append(BInitializer(lay_actf[-1], lay, seed))
        kernel_initializer.append(KInitializer(lay_actf[-1], lay, seed))
        f = get_activation(lay_actf[-1])
        w = kernel_initializer[-1].w0
        if len(lay_actf) == 1:
            activations.append(SciActivation(w, f))
        else:
            activations.append(SciActivationLayer(w, f))

    return activations, bias_initializer, kernel_initializer


def unpack_singleton(x):
    """Gets the first element if the iterable has only one value.

    Otherwise return the iterable.

    # Argument
        x: A list or tuple.

    # Returns
        The same iterable or the first element.
    """
    if len(x) == 1:
        return x[0]
    return x


def to_list(x, allow_tuple=False):
    """Normalizes a list/tensor into a list.

    If a tensor is passed, we return
    a list of size 1 containing the tensor.

    # Arguments
        x: target object to be normalized.
        allow_tuple: If False and x is a tuple,
            it will be converted into a list
            with a single element (the tuple).
            Else converts the tuple to a list.

    # Returns
        A list.
    """
    if isinstance(x, list):
        return x
    if allow_tuple and isinstance(x, tuple):
        return list(x)
    return [x]




'''
# Curve fitting in 1D

Here, a 1D curve fitting example is explored. Imagine, a synthetic data
generated from \\\( \sin(x) \\\) over the range of \\\( [0, 2\pi] \\\).

To train a neural network model on this curve, you should first define a `Variable`.

A neural network with three layers, each containing 10 neurons, and with `tanh` activation function is then generated
using the `Functional` class.

The target is imposed on the output using the `Data` class from `Constraint`, and passed to the `SciModel` to form a
Sciann model.
'''

import numpy as np
from sciann.functionals.rnn_variable import RNNVariable
from sciann.functionals.rnn_functional import RNNFunctional
from sciann.functionals.rnn_field import RNNField
from sciann import SciModel
from sciann.utils import diff, set_random_seed
from sciann.constraints import Data, Tie

set_random_seed(1234)

tunits = 3

# Synthetic data generated from sin function over [0, 2pi]
# x_true = np.linspace(0, np.pi*2, 10000).reshape(-1, tunits, 1)
x_true = np.linspace(0, np.pi*2, 100).reshape(-1, 1)
dx = np.diff(x_true.flatten()).mean()
x_true = x_true + np.linspace(0, dx*(tunits-1), tunits).reshape(1, -1)

y_true = np.sin(x_true)
dy_true = np.cos(x_true) #.sum(axis=1, keepdims=True)*2.0

# The network inputs should be defined with Variable.
t = RNNVariable(tunits, name='t', dtype='float64')

# Each network is defined by Functional.
y = RNNFunctional(
    'y', t, [5],
    activation='tanh',
    recurrent_activation='linear',
    kernel_initializer=1.0,
    recurrent_initializer=1.0,
    bias_initializer=1.0,
    rnn_type="SimpleRNN"
)
dy= diff(y, t)

def rnn_cell(x):
    y = []
    actf = lambda x: x
    for i, xi in enumerate(x):
        if i==0:
            y.append(
                actf(np.ones([1,5]) + xi*np.ones([1,5]))
            )
        else:
            y.append(
                actf(np.ones([1, 5]) + xi * np.ones([1, 5]) + np.matmul(y[-1], np.ones([5, 5])))
            )

    actf = lambda x: np.tanh(x)
    y = [actf(yi) for yi in y]
    return y

def rnn_net(x):
    actf = lambda x: x
    ys = rnn_cell(x)
    return np.concatenate([1 + np.matmul(actf(y), np.ones([5,1])) for y in ys], axis=-1).flatten()

test_rnn = lambda i: y.eval(x_true[i,:]) - rnn_net(x_true[i, :])

for i in range(10):
    print(test_rnn(i))

raise ValueError

# Define the target (output) of your model.
c1 = Data(y)
# c2 = Data((y[1]-y[0])/dx)
c2 = Data(dy)


# The model is formed with input `x` and condition `c1`.
model = SciModel(t, [c1, c2])

# Training: .train runs the optimization and finds the parameters.
model.train(
    x_true.reshape(-1, tunits, 1), 
    [y_true.reshape(-1, tunits, 1), dy_true.reshape(-1, tunits, 1)], 
    batch_size=1000000, epochs=20000, learning_rate=0.01
)

# used to evaluate the model after the training.
# x_pred = np.linspace(0, np.pi*4, 20000).reshape(-1, tunits, 1)
x_pred = np.linspace(0, np.pi*4, 200).reshape(-1, 1)
dx = np.diff(x_pred.flatten()).mean()
x_pred = x_pred + np.linspace(0, dx*(tunits-1), tunits).reshape(1, -1)

y_pred = y.eval(model, x_pred)
# dy_pred = dy.eval(model, x_pred)

y_star = np.sin(x_pred)
dy_star = np.cos(x_pred)

import matplotlib.pyplot as plt
plt.plot(x_pred[:,0], y_pred[:,0], x_pred[:,0], y_star[:,0])
# plt.plot(x_pred, dy_pred.reshape(-1), x_pred, dy_star)
plt.show()




import tensorflow.python.keras as k


class SciKernelInitializer(k.initializers.VarianceScaling):

    def __init__(self, actf='linear', lay=0, seed=None):
        self.actf = actf
        self.lay = lay

        # Kernel initializer
        if actf in ('linear', 'relu'):
            self.w0 = 1.0
            scale = 1.0
            distribution = 'truncated_normal'
            mode = 'fan_avg'
        elif actf in ('tanh', 'atan'):
            self.w0 = 1.0
            scale = 1.0
            distribution = 'truncated_normal'
            mode = 'fan_avg'
        elif actf in ('sin', 'cos'):
            self.w0 = 6.0 if lay==0 else 1.0
            scale = 1.0
            distribution = 'truncated_normal'
            mode = 'fan_in'
        else:
            self.w0 = 1.0
            scale = 1.0
            distribution = 'truncated_normal'
            mode = 'fan_avg'

        super(SciKernelInitializer, self).__init__(
            scale=scale,
            mode=mode,
            distribution=distribution,
            seed=seed
        )

    def get_config(self):
        base_config = super().get_config()
        config = {
            'w0': self.w0,
            'lay': self.lay,
            'actf': self.actf,
            'bias': self.bias
        }
        return dict(list(base_config.items()) + list(config.items()))


class SciBiasInitializer(k.initializers.RandomUniform):

    def __init__(self, actf='linear', lay=0, seed=None):
        self.w0 = 1.0
        self.actf = actf
        self.lay = lay

        super(SciBiasInitializer, self).__init__(seed=seed)

    def get_config(self):
        base_config = super().get_config()
        config = {
            'w0': self.w0,
            'lay': self.lay,
            'actf': self.actf
        }
        return dict(list(base_config.items()) + list(config.items()))


class SciOrthogonalInitializer(k.initializers.Orthogonal):

    def __init__(self, actf='linear', lay=0, bias=False, seed=None):
        self.actf = actf
        self.bias = bias

        if actf in ('linear', 'relu'):
            self.w0 = 1.0
            scale = 2.0
        elif actf in ('tanh', 'atan'):
            self.w0 = 1.0
            scale = 1.0
        elif actf in ('sin', 'cos'):
            self.w0 = 30.0 if lay==0 else 1.0
            scale = 2.0
        else:
            self.w0 = 1.0
            scale = 1.0

        if bias is True:
            self.w0 = 1./self.w0

        super(SciOrthogonalInitializer, self).__init__(
            gain=scale,
            seed=seed
        )

    def get_config(self):
        base_config = super().get_config()
        config = {
            'w0': self.w0,
            'actf': self.actf,
            'bias': self.bias
        }
        return dict(list(base_config.items()) + list(config.items()))


k.utils.generic_utils.get_custom_objects().update({
    'SciKernelInitializer': SciKernelInitializer,
    'SciOrthogonalInitializer': SciOrthogonalInitializer
})



""" Tie constraint to tie different outputs of the network.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .constraint import Constraint
from ..utils import is_functional


class Tie(Constraint):
    """ Tie class to constrain network outputs.
        constraint: `cond1 - cond2 == sol`.

    # Arguments
        cond1: Functional.
            A `Functional` object to be tied to cond2.
        cond2: Functional.
            A 'Functional' object to be tied to cond1.
        name: String.
            A `str` for name of the pde.

    # Returns

    # Raises
        ValueError: 'pde' should be a functional object.
    """
    def __init__(self, cond1, cond2, name="tie", **kwargs):
        # prepare cond.
        if not is_functional(cond1):
            raise ValueError(
                "Expected a Functional object as the cond1, received a "
                "{} - {}".format(type(cond1), cond1)
            )
        if not is_functional(cond2):
            raise ValueError(
                "Expected a Functional object as the cond2, received a "
                "{} - {}".format(type(cond2), cond2)
            )
        # Form the constraint.
        try:
            cond = cond1-cond2
        except (ValueError, TypeError):
            print(
                'Unexpected ValueError/TypeError - ',
                'make sure `cond1` and `cond2` are functional objects. \n',
                'cond1 - {} \n'.format(cond1),
                'cond2 - {} \n'.format(cond2)
            )
        # Check inputs.
        if 'mesh_ids' in kwargs or 'sol' in kwargs:
            raise ValueError(
                'Legacy interface: please use `SciModel` and `SciModel.train` to impose on specific ids. '
            )
        elif len(kwargs)>0:
            raise ValueError(
                'Unrecognized input variable: {}'.format(kwargs.keys())
            )
        super(Tie, self).__init__(
            cond=cond,
            name=name
        )



import numpy as np
from sciann import Variable, Field, Functional, SciModel
from sciann.utils.math import diag_grad, div_grad, grad, dot
from sciann.utils.math import diag_part, diag
import keras.backend as K

num_node = 1000
famil = np.arange(0, num_node).reshape(-1, 1) + np.array([-1, 0, 1])
famil[famil>999] = num_node - famil[famil>999]

xtrain = np.linspace(-2*np.pi, 2*np.pi, num_node)
gf00 = np.concatenate([np.ones((num_node,1))*0.2, np.ones((num_node,1))*0.6, np.ones((num_node,1))*0.2], axis=-1)

x = Variable("x", units=3, dtype='float64')
GF00 = Variable("GF00", units=3, dtype='float64')
y = Field("y", units=3, dtype='float64')
fx = Functional(y, x, 4*[10], 'tanh')

yf = dot(fx, GF00)
gf = grad(fx, x)
gf1 = diag_grad(fx, x)
gf2 = div_grad(fx, x)

dotgf = dot(gf, diag(GF00))

raise ValueError


model = SciModel([x, GF00], gf)

model.train([xtrain[famil], gf00], np.cos(xtrain), epochs=1000)

xtest = np.linspace(-4*np.pi, 4*np.pi, num_node)
ypred = yf.eval(model, [xtest[famil], gf00])
dypred = gf.eval(model, [xtest[famil], gf00])

import matplotlib.pyplot as plt
plt.plot(xtest, ypred, 'b', xtest, np.sin(xtest), '--g')
plt.plot(xtest, dypred, 'r', xtest, np.cos(xtest), '--k')
plt.show()
